; AutoSampler version 3.0
; Written by Jeanette C., find the latest version on:
; http://juliencoder.de
; This software is licensed under the GPL (GNU General Public License)
; version 3 (see COPYING for details).
; Csound version: 6.15 (6.14 should work or GIT version later than January 2020)

; Auto Sampler allows you to automatically sample a sound from a hardware
; instrument. Tis "simple" sampler works best for sounds that decay naturally.
; There are two autosamplers to choose: AutoSampler (completely automatic)
; and AutoSamplerRetune, which will pause after a given number of notes,
; so you can retune your instrument, ideal for analogue synthesizers with
; unstable keytracking.

; To run this successfully you need to set your appropriate audio and MIDI
; devices. See in the CsOptions section directly below.
; Once you've set up everything simply run it with:
; csound autosampler.csd
<CsoundSynthesizer>
<CsOptions>
; Choose your realtime audio module. Remove the semicolon from the beginning
; of the correct line (ONLY from the beginning!!!)
;-+rtaudio=PortAudio ; All platforms
;-+rtaudio=alsa ; Linux: ALSA (Advanced Linux Sound Architecture)
;-+rtaudio=jack ; Linux JACK Audio Connection Kit
;-+rtaudio=mme ; Windows MultiMedia Extension (the Windows standard?)
;-+rtaudio=CoreAudio ; Mac OS standard

; Choose your realtime MIDI module. Rem ove the semicolon from the beginning
; of the correct line
;-+rtmidi=PortMIDI ; All platforms
;-+rtmidi=alsa ; Linux (Advanced Linux Sound Architecture)
;-+rtmidi=cmidi ; Mac OS (CoreMidi?)
;-+rtmidi=mme ; Windows MultiMedia Extension
;-+rtmidi=winmm ; Windows

; Choose your MIDI output port
; ALSA will use the hw:card,subdevice format, you can list your ports
; with amidi -l
; if possible this will also show proper names, especially with USB MIDI
; devices.
; All other MIDI modules only use one number, check your desktop multimedia
; control centre, audio-MIDI setup or whereever you can see your devices
; (maybe inside your DAW, if you use one).
;-Q hw:1,0 ; ALSA format)
;-Q 1 ; format for all others
; Leave these options and go down to sr and nchnls
-d --messagelevel=2 -Ma -o dac -i adc
</CsOptions>
; ==============================================
<CsInstruments>
; sr sets the samplerate, usually either 44100 or 48000Hz.
sr	=	48000
ksmps	=	1
; nchnls tells Csound how many audio channels your soundcard has.
; 2 should work for every soundcard, but you may have more (18 is excessive!)
nchnls	=	18
; Go down to the very bottom to setup the sampler (how many notes, velocity
; layers, etc.)
0dbfs	=	1
massign 0, 0

; Shut down Csound when the work is done
instr Exit
	exitnow
endin

; Instrument for retuning
instr Retune
	; Initialise parameters and controls
	kStatus init 0
	kChan init 0
	kData1 init 48
	kData2 init 0
	kInitialised init 0 ; 1 when the tuning note is set
	kFreq init cpsmidinn(48) ; frequency derived from MIDI note
	kDone init 0 ; set to 1 when finished
	kKey init 0 ; keyboard input
	kPressed init 0 ; second output of sensekey
	kVal1 init 1 ; for chnset
	
	; Produce the reference tone
	aReference = vco2(.05, kFreq)

	; Print instructions
	prints("\n----------------------------------------\nRetune your synthesizer:")
	prints("\nPlay a note, until you find the right pirch, then press 's' on your computer")
	prints("\nkeyboard and tune your synthesizer to the reference tone.\n")

	; Command interface, read keyboard commands
	kKey, kPressed sensekey
	if (kPressed == 1) then ; if a key is pressed
		if ((kKey == 83) || (kKey == 115)) then ; s for retune start
			kInitialised = 1
		elseif ((kKey == 81) || (kKey == 113)) then ; q for quit/done
			kDone = 1
			chnset(kVal1, "AutoSampler.Tuned")
		endif ; key switching
	endif ; key was pressed

	if (kInitialised == 0) then ; find a note
		kStatus, kChan, kData1, kData2 midiin
		kFreq = cpsmidinn(kData1)
	endif

	if (changed2(kInitialised) != 0) then
		printks("\nNow tune your synthesizer and then press 'q' on your computer", 1/kr)
		printks("\nkeyboard to continue sampling.", 1/kr)
	endif

	if (kDone == 1) then
		turnoff
	endif

	; Output the reference tone
	outs(aReference, aReference)
endin

; Record a single note at given velocity and length for the specified round
; robin index
; p4 = iMidiNote, p5 = iVelocity, p6 = iRoundRobin, p7 = iLength
; p8 = iCurLayer
instr Record
	; Initialise the parameters
	iNote init p4
	iVelocity init p5
	iRoundRobin init p6
	iLength init p7
	iCurLayer init p8
	Sname chnget "AutoSampler.Name"
	iStereo = chnget("AutoSampler.Stereo")
	iMidiChannel = chnget("AutoSampler.Channel")

	; Controls for realtime
	kStartTrig init 1
	kStopTrig init 0
	kStopTrig vdelayk kStartTrig, iLength, (iLength + .1)
	kRelease init 0
	kRelStart init 0
	kNow times
	kDone init 0

	; Terminate instrument when done
	if (kDone == 1) then
		turnoff
	endif

	; Create filename
	Sfilename = sprintf("Samples/%s-%d-%d-%d.wav", Sname, iNote, iVelocity, iRoundRobin)

	; Set up audio signals and record
	aLeft init 0
	aRight init 0

	if (iStereo == 0) then
		aLeft = chnget:a("AutoSampler.Left")
		kRms1 = rms(aLeft)
		kRms2 init 0
		fout(Sfilename, -1, aLeft)
	else
		aLeft = chnget:a("AutoSampler.Left")
		aRight = chnget:a("AutoSampler.Right")
		kRms1 = rms(aLeft)
		kRms2 = rms(aRight)
		fout(Sfilename, -1, aLeft, aRight)
	endif

	; trigger MIDI, stop MIDI and wait for audio tail to decay
	if (kRelease == 0) then
		if (kStartTrig == 1) then ; Start note
			midiout(144, iMidiChannel, iNote, iVelocity)
		endif
		if (kStopTrig == 1) then ; Stop note
			midiout(128, iMidiChannel, iNote, iVelocity)
			kRelease = 1
			kRelStart times
		endif
	else ; Allow audio to decay
		if ((kRms1 <.001) && (kRms2 <.001)) then
			kDone = 1
			chnset(kDone, "AutoSampler.Ready")
			schedulek("MakeSFZ", 0, .1, iNote, iVelocity, iRoundRobin, (kNow - kRelStart), iCurLayer)
		endif
	endif

	kStartTrig = 0
endin

; Print the SFZ header
instr MakeHeader
	Sname chnget "AutoSampler.Name"
	Sfilename = strcat(Sname, ".sfz")

	fprints(Sfilename, "// SFZ file generated by Jeanette C's AutoSampler, powered by Csound\n")
	fprints(Sfilename, "//==============================================================================\n")
	turnoff
endin

; Write an SFZ entry to the file
; p4 = iNote, p5 = iVelocity, p6 = iRobin, p7 = iRelease, p8 = iCurLayer
instr MakeSFZ
	; Initialise the parameters
	iNote init p4
	iVelocity init p5
	iRobin init p6
	iReleaseTime init p7
	iCurLayer init p8
	Sname chnget "AutoSampler.Name"

	; Get additional data from channels
	iRobinStep = chnget:i("AutoSampler.RobinStep")
	iDownRange = chnget:i("AutoSampler.Down")
	iUpRange = chnget:i("AutoSampler.Up")
	iVelLayers = chnget:i("AutoSampler.Layers")

	; Calculate values for the SFZ file
	if (iNote != chnget:i("AutoSampler.First")) then
		iLokey = iNote - iDownRange
	else
		iLokey = iNote
	endif

	if (iNote != chnget:i("AutoSampler.Last")) then
		iHikey = iNote + iUpRange
	else
		iHikey = iNote
	endif

	if ((iCurLayer + 1) == iVelLayers) then
		iLovel = 1
	else
		iLovel = int(127 * (1 - ((iCurLayer + 1) / iVelLayers)))
	endif

	iLorand = (iRobin -1) * iRobinStep
	iHirand = iRobin * iRobinStep

	; Create SFZ filename
	Sfilename = strcat(Sname, ".sfz")

	; Write data to SFZ file
	if ((iVelocity == 127) && (iRobin == 1)) then
		fprints(Sfilename, "\n<group> amp_veltrack=0\n")
	endif

	fprints(Sfilename,"<region>\n")
	fprints(Sfilename,"sample=Samples\\\\%s-%d-%d-%d.wav\n", Sname, iNote, iVelocity, iRobin)
	fprints(Sfilename,"ampeg_release=%f\n", iReleaseTime)
	fprints(Sfilename,"lokey=%d\nhikey=%d\npitch_keycenter=%d\n", iLokey, iHikey, iNote)
	fprints(Sfilename,"lovel=%d\nhivel=%d\n", iLovel, iVelocity)
	if (iRobinStep != 1) then ; we have round robins
		fprints(Sfilename,"lorand=%f\nhirand=%f\n", iLorand, iHirand)
	endif
	fprints(Sfilename, "\n")
	turnoff
endin

; The master instrument to auto sample
; p4 = SinstrumentName, p5 = iNoteLength, p6 = iFirstNote, p7 = iLastNote,
; p8 = iNoteStep, p9 = iVelocityLayers, p10 = iRoundRobins
; p11 = iLeftChannel, p12 = iRightChannel(-1 for mono), p13 = iMidiChannel,
; p14 = iRetuneInterval
instr AutoSamplerRetune
	Sname = strget(p4)
	iNoteLength init p5
	iFirstNote init p6
	iLastNote init p7
	iNoteStep init p8
	iVelocityLayers init p9
	iRoundRobins init p10
	iLeftChannel init p11
	iRightChannel init p12
	iMidiChannel init p13
	iRetuneInterval init p14
	kVal0 init 0 ; just a 0 for chnset at k-rate
	kRetune init 0 ; set to 1 if retuning is necessary
	kRetuneCount init 0 ; notes since last retune

	; Set global channels
	chn_S("AutoSampler.Name", 3)
	chnset(Sname, "AutoSampler.Name")
	chn_k("AutoSampler.Channel", 3)
	chnset(iMidiChannel, "AutoSampler.Channel")
	chn_k("AutoSampler.Ready", 3)
	chnset(1, "AutoSampler.Ready") ; needs to be ready for first sampling
	chn_k("AutoSampler.Stereo", 3)
	chn_k("AutoSampler.First", 3)
	chnset(iFirstNote, "AutoSampler.First")
	chn_k("AutoSampler.Last", 3)
	chnset(iLastNote, "AutoSampler.Last")
	chn_k("AutoSampler.Layers", 3)
	chnset(iVelocityLayers, "AutoSampler.Layers")
	chn_k("AutoSampler.RobinStep", 3)
	chnset((1 / iRoundRobins), "AutoSampler.RobinStep")
	iDownRange = int(iNoteStep / 2)
	chn_k("AutoSampler.Down", 3)
	chnset(iDownRange, "AutoSampler.Down")
	chn_k("AutoSampler.Up", 3)
	chnset((iNoteStep - iDownRange -1), "AutoSampler.Up")
	chn_k("AutoSampler.Tuned", 3)
	chnset(0, "AutoSampler.Tuned")

	if (iRightChannel == -1) then
		chnset(0, "AutoSampler.Stereo")
	else
		chnset(1, "AutoSampler.Stereo")
	endif

	; Create Samples folder
	iMkdir = system_i(1,"mkdir Samples")

	; Write SFZ file header
	schedule("MakeHeader", 0, .1)

	; Input audio and copy to channels
	aLeft init 0
	aRight init 0

	if (k(iRightChannel) == -1) then
		aLeft inch iLeftChannel
		chnset(aLeft, "AutoSampler.Left")
	else
		aLeft, aRight inch iLeftChannel, iRightChannel
		chnset(aLeft, "AutoSampler.Left")
		chnset(aRight, "AutoSampler.Right")
	endif

	; Set up variables for looping
	kCurNote init iFirstNote
	kCurRobin init 1 ; This is encoded into the filename and counted from 1
	kCurVelocity init 127
	kCurLayer init 0

	; Record notes
	if (chnget:k("AutoSampler.Ready") == 1) then
		if ((changed2(kRetune) != 0) && (kRetune == 1)) then ; if retune has been set
			schedulek("Retune", 0, -1)
		endif

		if (kRetune == 0) then ; record a note
			chnset(kVal0, "AutoSampler.Ready")
			if (kCurNote <= iLastNote) then
				printks("\nRecording note %d, round robin %d velocity %d", 1/kr, kCurNote, kCurRobin, kCurVelocity)
				schedulek("Record", 0, -1, kCurNote, kCurVelocity, kCurRobin, iNoteLength, kCurLayer)
				if (kCurRobin < iRoundRobins) then
					kCurRobin += 1
				else ; check for velocity next
					kCurRobin = 1
					if (kCurLayer < (iVelocityLayers - 1)) then ; Next Velocity
						kCurLayer += 1
						kCurVelocity = int(127 * (1 - (kCurLayer / iVelocityLayers)))
					else ; Next note
						kCurLayer = 0
						kCurVelocity = 127
						kCurNote += iNoteStep
						kRetuneCount += iNoteStep
						if (kRetuneCount >= iRetuneInterval) then
							kRetune = 1
							kRetuneCount = 0
						endif
					endif
				endif
			else ; kCurNote > iLastNote: stop
				schedulek("Exit", 0, .1)
			endif
		else ; retuning
			if (chnget:k("AutoSampler.Tuned") == 1) then
				kRetune = 0
				chnset(kVal0, "AutoSampler.Tuned")
			endif
		endif ; if retuning
	endif ; if ready for next note

	; Output the audio live
	if (iRightChannel == -1) then
		outs(aLeft, aLeft)
	else
		outs(aLeft, aRight)
	endif
endin

; Auto sampler without retuning, a wrapper for AutoSamplerRetune
; p4 = SinstrumentName, p5 = iNoteLength, p6 = iFirstNote, p7 = iLastNote
; p8 = iNoteStep, p9 = iVelocityLayers, p10 = iRoundRobins
; p11 = iLeftChannel, p12 = iRightChannel (-1 for mono), p13 = iMidiChannel
instr AutoSampler
	; Initialise parameters
	Sname = strget(p4)
	schedule("AutoSamplerRetune", 0, p3, Sname, p5, p6, p7, p8, p9, p10, p11, \
		p12, p13, 200)
	turnoff
endin

</CsInstruments>
; ==============================================
<CsScore>
; Remember: right audio channel = -1 for mono
; AutoSampler simple
; p4=Sname p5=iNoteLength, p6=iFirstNote p7=iLastNote p8=iNoteInterval
; p9=iVelocityLayers p10=iRoundRobins 
; p11=iLeftAudioChannel p12=iRightAudioChannel p13=iMidiChannel
; Example: Sample the instrument Guitar, play each note for 1.0 seconds,
; from MIDI note 36 to MIDI note 96, sample every third note, record five
; velocity layers and four round robins at each layer, record audio on channels
; 7 and 8 and play notes on MIDI channel 1.
; Modify the line below, but don't forget to remove the semicolon at the
; beginning.
;i"AutoSampler" 0 z "Guitar" 1.0 36 96 3 5 4 7 8 1

; Auto Sampler with retuning pauses
; p4=Sname p5=iNoteLength, p6=iFirstNote p7=iLastNote p8=iNoteInterval
; p9=iVelocityLayers p10=iRoundRobins 
; p11=iLeftAudioChannel p12=iRightAudioChannel p13=iMidiChannel
; p14=iRetuneInterval
; Example: sample instrument Saw, hold each note for .2 seconds, sample from 
; note 36 to note 96, sample every 3rd key, four velocity layers, two round
; robins on each velocity, record audio from channels 3 and 4 and MIDI channel
; 1, and pause every twelve notes (one octave) so you can retune the synth
; Modify the line below, but don't forget to remove the semicolon at the
; beginning.
;i"AutoSamplerRetune" 0 z "Brute" .2 36 96 3 4 2 3 4 1 12
e
</CsScore>
</CsoundSynthesizer>
